@@ -135,12 +135,6 @@ void reset_response_body(client_t * client)
 
 // ============== request processing ==================================================
 
-int on_reset(llhttp_t * parser)
-{
-    LOGf("%s: Detect HTTP pipelining! Pipelining not supported!", __func__);
-    return -1;
-}
-
 int on_message_begin(llhttp_t * parser)
 {
     LOGi("on_message_begin: ------------------------------");
@@ -427,7 +421,7 @@ int on_message_complete(llhttp_t * parser)
     }
     if (client->error) {
         if (client->request.expect_continue && client->error == HTTP_STATUS_EXPECTATION_FAILED)
-            return 0;
+            return HPE_PAUSED;
         return -1;
     }
 
@@ -436,7 +430,7 @@ int on_message_complete(llhttp_t * parser)
         const int64_t ilen = client->request.wsgi_input_size;
         if (ilen != clen) {
             client->error = 1;
-            LOGc("Received body with size %lld not equal specified 'Content-Length' = %lld", ilen, clen);
+            LOGc("Received body with size %lld not equal specified 'Content-Length' = %lld", (long long)ilen, (long long)clen);
             return -1;
         }
     }
@@ -475,13 +469,20 @@ int on_message_complete(llhttp_t * parser)
         set_header(client, g_cv.REMOTE_ADDR, client->remote_addr, -1, 0);
 
     if (client->request.chunked && client->request.http_content_length < 0) {
-        sprintf(buf, "%lld", client->request.wsgi_input_size);
+        sprintf(buf, "%lld", (long long)client->request.wsgi_input_size);
         set_header(client, g_cv.CONTENT_LENGTH, buf, -1, 0);
         // Insertion of this header is allowed because the header "Transfer-Encoding" FastWSGI server removes!
         // https://peps.python.org/pep-3333/#other-http-features
     }
 
     client->request.load_state = LS_OK;
+    return HPE_PAUSED;
+}
+
+int on_reset(llhttp_t * parser)
+{
+    // This callback called only on HTTP pipelining
+    //LOGi("%s: Detect HTTP pipelining!", __func__);
     return 0;
 }
 
@@ -623,7 +624,7 @@ int build_response(client_t * client, int flags, int status, const void * header
     }
     if (response) {
         char scode[4];
-        size_t status_len = 0;
+        Py_ssize_t status_len = 0;
         const char * status_code = PyUnicode_AsUTF8AndSize(response->status, &status_len);
         if (status_len < 3)
             return -2;
@@ -648,7 +649,7 @@ int build_response(client_t * client, int flags, int status, const void * header
         for (Py_ssize_t i = 0; i < hsize; i++) {
             PyObject* tuple = PyList_GET_ITEM(response->headers, i);
 
-            size_t key_len = 0;
+            Py_ssize_t key_len = 0;
             const char * key = PyUnicode_AsUTF8AndSize(PyTuple_GET_ITEM(tuple, 0), &key_len);
 
             if (key_len == 14 && key[7] == '-')
@@ -663,7 +664,7 @@ int build_response(client_t * client, int flags, int status, const void * header
                 if (strcasecmp(key, "Date") == 0)
                     resp_date_present = 1;
 
-            size_t value_len = 0;
+            Py_ssize_t value_len = 0;
             const char * value = PyUnicode_AsUTF8AndSize(PyTuple_GET_ITEM(tuple, 1), &value_len);
 
             xbuf_add(head, key, key_len);
@@ -730,8 +731,8 @@ end:
     }
     else if (body_size > 0) {
         char * buf = xbuf_expand(head, 48);
-        head->size += sprintf(buf, "Content-Length: %lld\r\n", body_size);
-        LOGi("Added Header 'Content-Length: %lld'", body_size);
+        head->size += sprintf(buf, "Content-Length: %lld\r\n", (long long)body_size);
+        LOGi("Added Header 'Content-Length: %lld'", (long long)body_size);
     }
     xbuf_add(head, "\r\n", 2);  // end of headers
 
@@ -751,9 +752,9 @@ int get_info_from_wsgi_response(client_t * client)
     Py_ssize_t hsize = PyList_GET_SIZE(response->headers);
     for (Py_ssize_t i = 0; i < hsize; i++) {
         PyObject* tuple = PyList_GET_ITEM(response->headers, i);
-        size_t key_len = 0;
+        Py_ssize_t key_len = 0;
         const char * key = PyUnicode_AsUTF8AndSize(PyTuple_GET_ITEM(tuple, 0), &key_len);
-        size_t value_len = 0;
+        Py_ssize_t value_len = 0;
         const char * value = PyUnicode_AsUTF8AndSize(PyTuple_GET_ITEM(tuple, 1), &value_len);
         if (key_len == 14 && key[7] == '-' && strcasecmp(key, "Content-Length") == 0) {
             if (value_len == 0)
@@ -766,7 +767,7 @@ int get_info_from_wsgi_response(client_t * client)
                 if (clen <= 0 || clen == LLONG_MAX)
                     return -3;  // error
             }
-            LOGi("wsgi response: content-length = %lld", clen);
+            LOGi("wsgi response: content-length = %lld", (long long)clen);
             client->response.wsgi_content_length = clen;
         }
     }
@@ -794,7 +795,7 @@ PyObject* wsgi_iterator_get_next_chunk(client_t * client, int outpyerr)
         return NULL;
     }
     PyObject* item;
-    while (item = PyIter_Next(client->response.body_iterator)) {
+    while ( (item = PyIter_Next(client->response.body_iterator)) != NULL ) {
         if (!PyBytes_Check(item)) {
             client->error = 11;
             if (outpyerr)
@@ -874,7 +875,7 @@ int wsgi_body_pleload(client_t * client, PyObject * wsgi_body)
     int fully_loaded = 1;
     ssize_t prev_size = 0;
     size_t chunks = 0;
-    while (item = wsgi_iterator_get_next_chunk(client, 1)) {
+    while ( (item = wsgi_iterator_get_next_chunk(client, 1)) != NULL ) {
         ssize_t item_size = PyBytes_GET_SIZE(item);
         body[chunks++] = item;
         client->response.body_chunk_num = chunks;
@@ -889,7 +890,7 @@ int wsgi_body_pleload(client_t * client, PyObject * wsgi_body)
         }
         if (is_filelike) {
             // Only filelike data source check for max chunk size
-            if (client->response.body_preloaded_size + prev_size >= g_srv.max_chunk_size) {
+            if (client->response.body_preloaded_size + prev_size >= (int64_t)g_srv.max_chunk_size) {
                 LOGd("wsgi_body: overflow chunk buffer! Preloaded size = %d", client->response.body_preloaded_size);
                 fully_loaded = 0;
                 break;
@@ -906,17 +907,17 @@ int wsgi_body_pleload(client_t * client, PyObject * wsgi_body)
             LOGc("wsgi_body: real size of the response body exceeds the specified Content-Length");
             return -14;
         }
-        LOGi("wsgi_body: response body fully loaded! (size = %lld)", client->response.body_preloaded_size);
+        LOGi("wsgi_body: response body fully loaded! (size = %lld)", (long long)client->response.body_preloaded_size);
         client->response.body_total_size = client->response.body_preloaded_size;
         return 1;
     }
     if (client->response.body_preloaded_size == wsgi_content_length) {
-        LOGi("wsgi_body: response body fully loaded! (SIZE = %lld)", wsgi_content_length);
+        LOGi("wsgi_body: response body fully loaded! (SIZE = %lld)", (long long)wsgi_content_length);
         client->response.body_total_size = wsgi_content_length;
         return 1;
     }
     if (wsgi_content_length > 0) {
-        LOGd("wsgi_body: large content transfer begins. body_total_size = %lld", wsgi_content_length);
+        LOGd("wsgi_body: large content transfer begins. body_total_size = %lld", (long long)wsgi_content_length);
         client->response.body_total_size = wsgi_content_length;
         return 1;
     }
@@ -953,7 +954,6 @@ void init_request_dict()
 void configure_parser_settings(llhttp_settings_t * ps)
 {
     llhttp_settings_init(ps);
-    ps->on_reset = on_reset;
     ps->on_message_begin = on_message_begin;
     ps->on_url = on_url;
     ps->on_url_complete = on_url_complete;
@@ -964,4 +964,5 @@ void configure_parser_settings(llhttp_settings_t * ps)
     ps->on_headers_complete = on_headers_complete;
     ps->on_body = on_body;
     ps->on_message_complete = on_message_complete;
+    ps->on_reset = on_reset;
 }
